@@ -22,15 +22,26 @@
 
 #define PY_SSIZE_T_CLEAN
 
+/* Include order matters! */
+#include "Python.h"
+
 /* Search lib/win32 first, then fallthrough to <stdint.h> as required.*/
 #include "stdint.h"
 
 #include <errno.h>
 #include <stdarg.h>
 #include <string.h>
+
+#ifdef _WIN32
+#   define bool int
+#   define true 1
+#   define false 0
+#else
+#   include <stdbool.h>
+#endif
+
 #include <sys/stat.h>
 
-#include "Python.h"
 #include "structmember.h"
 
 #ifdef HAVE_MEMSINK
@@ -43,6 +54,7 @@
 #endif
 
 #include "lmdb.h"
+#include "preload.h"
 
 
 /* Comment out for copious debug. */
@@ -144,22 +156,6 @@ typedef struct TransObject TransObject;
 
 #endif
 
-#if (PY_MAJOR_VERSION == 2) && (PY_MINOR_VERSION < 6)
-static PyObject *
-PyUnicode_FromString(const char *u)
-{
-    PyObject *s = PyString_FromString(u);
-    if(s) {
-        PyObject *t = PyUnicode_FromEncodedObject(
-            s, Py_FileSystemDefaultEncoding, "strict");
-        Py_DECREF(s);
-        s = t;
-    }
-    return s;
-}
-#endif
-
-
 struct list_head {
     struct lmdb_object *prev;
     struct lmdb_object *next;
@@ -228,6 +224,8 @@ enum trans_flags {
 /** lmdb.Transaction */
 struct TransObject {
     LmdbObject_HEAD
+    /** Python-managed list of weakrefs to this object. */
+    PyObject *weaklist;
     EnvObject *env;
 #ifdef HAVE_MEMSINK
     /** Copy-on-invalid list head. */
@@ -388,7 +386,8 @@ static const struct error_map error_map[] = {
     {MDB_PAGE_FULL, "PageFullError"},
     {MDB_MAP_RESIZED, "MapResizedError"},
     {MDB_INCOMPATIBLE, "IncompatibleError"},
-    {MDB_BAD_RSLOT, "BadRSlotError"},
+    {MDB_BAD_RSLOT, "BadRslotError"},
+    {MDB_BAD_DBI, "BadDbiError"},
     {MDB_BAD_TXN, "BadTxnError"},
     {MDB_BAD_VALSIZE, "BadValsizeError"},
     {EACCES, "ReadonlyError"},
@@ -594,7 +593,6 @@ val_from_buffer(MDB_val *val, PyObject *buf)
         (Py_ssize_t *) &val->mv_size);
 }
 
-
 /* ------------------- */
 /* Concurrency control */
 /* ------------------- */
@@ -604,6 +602,10 @@ val_from_buffer(MDB_val *val, PyObject *buf)
     out = (e); \
     Py_END_ALLOW_THREADS
 
+#define PRELOAD_UNLOCKED(_rc, _data, _size) \
+    Py_BEGIN_ALLOW_THREADS \
+    preload(_rc, _data, _size); \
+    Py_END_ALLOW_THREADS
 
 /* ---------------- */
 /* Argument parsing */
@@ -685,7 +687,7 @@ parse_arg(const struct argspec *spec, PyObject *val, void *out)
             *((PyObject **) dst) = val;
             break;
         case ARG_BOOL:
-            *((int *)dst) = val == Py_True;
+            *((int *)dst) = PyObject_IsTrue(val);
             break;
         case ARG_BUF:
             ret = val_from_buffer((MDB_val *)dst, val);
@@ -755,7 +757,7 @@ parse_args(int valid, int specsize, const struct argspec *argspec,
     }
 
     if(args) {
-        int size = (int) PyTuple_GET_SIZE(args);
+        Py_ssize_t size = PyTuple_GET_SIZE(args);
         if(size > specsize) {
             type_error("too many positional arguments.");
             return -1;
@@ -791,8 +793,7 @@ parse_args(int valid, int specsize, const struct argspec *argspec,
 
             i = READ_ID(specidx);
             if(set & (1 << i)) {
-                PyErr_Format(PyExc_TypeError, "duplicate argument: %s",
-                             PyBytes_AS_STRING(pkey));
+                PyErr_Format(PyExc_TypeError, "duplicate argument: %U", pkey);
                 return -1;
             }
 
@@ -860,36 +861,40 @@ make_trans(EnvObject *env, DbObject *db, TransObject *parent, int write, int buf
         self = env->spare_txns;
         DEBUG("found freelist txn; self=%p self->txn=%p", self, self->txn)
         env->spare_txns = self->spare_next;
-        env->max_spare_txns++;
-        self->flags &= ~TRANS_SPARE;
-        _Py_NewReference(self);
-        UNLOCKED(rc, mdb_txn_renew(self->txn));
 
+        UNLOCKED(rc, mdb_txn_renew(self->txn));
         if(rc) {
             mdb_txn_abort(self->txn);
-            self->txn = NULL;
+            PyObject_Del(self);
+            return err_set("mdb_txn_begin", rc);
         }
+
+        env->max_spare_txns++;
+        self->flags &= ~TRANS_SPARE;
+
     } else {
+        MDB_txn *txn;
         if(write && env->readonly) {
             const char *msg = "Cannot start write transaction with read-only env";
             return err_set(msg, EACCES);
         }
 
-        if(! ((self = PyObject_New(TransObject, &PyTransaction_Type)))) {
-            return NULL;
+        flags = write ? 0 : MDB_RDONLY;
+        UNLOCKED(rc, mdb_txn_begin(env->env, parent_txn, flags, &txn));
+        if(rc) {
+            return err_set("mdb_txn_begin", rc);
         }
 
-        flags = write ? 0 : MDB_RDONLY;
-        UNLOCKED(rc, mdb_txn_begin(env->env, parent_txn, flags, &self->txn));
-    }
-
-    if(rc) {
-        PyObject_Del(self);
-        return err_set("mdb_txn_begin", rc);
+        if(! ((self = PyObject_New(TransObject, &PyTransaction_Type)))) {
+            mdb_txn_abort(txn);
+            return NULL;
+        }
+        self->txn = txn;
     }
 
     OBJECT_INIT(self)
     LINK_CHILD(env, self)
+    self->weaklist = NULL;
     self->env = env;
     Py_INCREF(env);
     self->db = db;
@@ -914,6 +919,7 @@ static PyObject *
 make_cursor(DbObject *db, TransObject *trans)
 {
     CursorObject *self;
+    MDB_cursor *curs;
     int rc;
 
     if(! trans->valid) {
@@ -925,19 +931,26 @@ make_cursor(DbObject *db, TransObject *trans)
         return NULL;
     }
 
-    self = PyObject_New(CursorObject, &PyCursor_Type);
-    UNLOCKED(rc, mdb_cursor_open(trans->txn, db->dbi, &self->curs));
+    UNLOCKED(rc, mdb_cursor_open(trans->txn, db->dbi, &curs));
     if(rc) {
-        PyObject_Del(self);
         return err_set("mdb_cursor_open", rc);
     }
 
+    self = PyObject_New(CursorObject, &PyCursor_Type);
+    if (!self) {
+        mdb_cursor_close(curs);
+        return NULL;
+    }
+
     DEBUG("sizeof cursor = %d", (int) sizeof *self)
     OBJECT_INIT(self)
     LINK_CHILD(trans, self)
+    self->curs = curs;
     self->positioned = 0;
     self->key.mv_size = 0;
+    self->key.mv_data = NULL;
     self->val.mv_size = 0;
+    self->val.mv_data = NULL;
     self->trans = trans;
     self->last_mutation = trans->mutations;
     self->dbi_flags = db->flags;
@@ -1037,29 +1050,20 @@ db_flags(DbObject *self, PyObject *args, PyObject *kwds)
     PyObject *dct;
     unsigned int f;
 
-    struct db_flags {
-        TransObject *txn;
-    } arg = {NULL};
-
-    static const struct argspec argspec[] = {
-        {"txn", ARG_TRANS, OFFSET(db_flags, txn)}
-    };
-
-    static PyObject *cache = NULL;
-    if(parse_args(self->valid, SPECSIZE(), argspec, &cache, args, kwds, &arg)) {
-        return NULL;
-    }
-    if(! arg.txn) {
-        return type_error("'txn' argument required");
-    }
-    if(! arg.txn->valid) {
-        return err_invalid();
+    if (args) {
+        Py_ssize_t size = PyTuple_GET_SIZE(args);
+        if(size > 1) {
+            return type_error("too many positional arguments.");
+        }
     }
 
     dct = PyDict_New();
     f = self->flags;
     PyDict_SetItemString(dct, "reverse_key", py_bool(f & MDB_REVERSEKEY));
     PyDict_SetItemString(dct, "dupsort", py_bool(f & MDB_DUPSORT));
+    PyDict_SetItemString(dct, "integerkey", py_bool(f & MDB_INTEGERKEY));
+    PyDict_SetItemString(dct, "integerdup", py_bool(f & MDB_INTEGERDUP));
+    PyDict_SetItemString(dct, "dupfixed", py_bool(f & MDB_DUPFIXED));
     return dct;
 }
 
@@ -1121,6 +1125,7 @@ static int
 env_clear(EnvObject *self)
 {
     MDEBUG("killing env..")
+
     INVALIDATE(self)
     self->valid = 0;
     Py_CLEAR(self->main_db);
@@ -1150,6 +1155,11 @@ env_clear(EnvObject *self)
 static void
 env_dealloc(EnvObject *self)
 {
+    if(self->weaklist != NULL) {
+        MDEBUG("Clearing weaklist..")
+        PyObject_ClearWeakRefs((PyObject *) self);
+    }
+
     env_clear(self);
     PyObject_Del(self);
 }
@@ -1261,14 +1271,9 @@ env_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
     fspath = PyBytes_AS_STRING(fspath_obj);
 
     if(arg.create && arg.subdir && !arg.readonly) {
-        struct stat st;
-        errno = 0;
-        stat(fspath, &st);
-        if(errno == ENOENT) {
-            if(mkdir(fspath, arg.mode)) {
-                PyErr_SetFromErrnoWithFilename(PyExc_OSError, fspath);
-                goto fail;
-            }
+        if(mkdir(fspath, arg.mode) && errno != EEXIST) {
+            PyErr_SetFromErrnoWithFilename(PyExc_OSError, fspath);
+            goto fail;
         }
     }
 
@@ -1357,21 +1362,25 @@ env_begin(EnvObject *self, PyObject *args, PyObject *kwds)
  * Environment.copy()
  */
 static PyObject *
-env_copy(EnvObject *self, PyObject *args)
+env_copy(EnvObject *self, PyObject *args, PyObject *kwds)
 {
     struct env_copy {
         PyObject *path;
-    } arg = {NULL};
+        int compact;
+    } arg = {NULL, 0};
 
     static const struct argspec argspec[] = {
-        {"path", ARG_OBJ, OFFSET(env_copy, path)}
+        {"path", ARG_OBJ, OFFSET(env_copy, path)},
+        {"compact", ARG_BOOL, OFFSET(env_copy, compact)}
     };
 
     PyObject *fspath_obj;
+    const char *fspath_s;
+    int flags;
     int rc;
 
     static PyObject *cache = NULL;
-    if(parse_args(self->valid, SPECSIZE(), argspec, &cache, args, NULL, &arg)) {
+    if(parse_args(self->valid, SPECSIZE(), argspec, &cache, args, kwds, &arg)) {
         return NULL;
     }
     if(! arg.path) {
@@ -1381,10 +1390,12 @@ env_copy(EnvObject *self, PyObject *args)
         return NULL;
     }
 
-    UNLOCKED(rc, mdb_env_copy(self->env, PyBytes_AS_STRING(fspath_obj)));
+    fspath_s = PyBytes_AS_STRING(fspath_obj);
+    flags = arg.compact ? MDB_CP_COMPACT : 0;
+    UNLOCKED(rc, mdb_env_copy2(self->env, fspath_s, flags));
     Py_CLEAR(fspath_obj);
     if(rc) {
-        return err_set("mdb_env_copy", rc);
+        return err_set("mdb_env_copy2", rc);
     }
     Py_RETURN_NONE;
 }
@@ -1393,11 +1404,12 @@ env_copy(EnvObject *self, PyObject *args)
  * Environment.copyfd(fd)
  */
 static PyObject *
-env_copyfd(EnvObject *self, PyObject *args)
+env_copyfd(EnvObject *self, PyObject *args, PyObject *kwds)
 {
     struct env_copyfd {
         int fd;
-    } arg = {-1};
+        int compact;
+    } arg = {-1, 0};
     int rc;
 #ifdef _WIN32
     PyObject *temp;
@@ -1405,16 +1417,19 @@ env_copyfd(EnvObject *self, PyObject *args)
 #endif
 
     static const struct argspec argspec[] = {
-        {"fd", ARG_INT, OFFSET(env_copyfd, fd)}
+        {"fd", ARG_INT, OFFSET(env_copyfd, fd)},
+        {"compact", ARG_BOOL, OFFSET(env_copyfd, compact)}
     };
+    int flags;
 
     static PyObject *cache = NULL;
-    if(parse_args(self->valid, SPECSIZE(), argspec, &cache, args, NULL, &arg)) {
+    if(parse_args(self->valid, SPECSIZE(), argspec, &cache, args, kwds, &arg)) {
         return NULL;
     }
     if(arg.fd == -1) {
         return type_error("fd argument required");
     }
+    flags = arg.compact ? MDB_CP_COMPACT : 0;
 
 #ifdef _WIN32
     temp = PyObject_CallMethod(msvcrt, "get_osfhandle", "i", arg.fd);
@@ -1426,13 +1441,13 @@ env_copyfd(EnvObject *self, PyObject *args)
     if(PyErr_Occurred()) {
         return NULL;
     }
-    UNLOCKED(rc, mdb_env_copyfd(self->env, (HANDLE)handle));
+    UNLOCKED(rc, mdb_env_copyfd2(self->env, (HANDLE)handle, flags));
 #else
-    UNLOCKED(rc, mdb_env_copyfd(self->env, arg.fd));
+    UNLOCKED(rc, mdb_env_copyfd2(self->env, arg.fd, flags));
 #endif
 
     if(rc) {
-        return err_set("mdb_env_copyfd", rc);
+        return err_set("mdb_env_copyfd2", rc);
     }
     Py_RETURN_NONE;
 }
@@ -1543,6 +1558,9 @@ env_open_db(EnvObject *self, PyObject *args, PyObject *kwds)
         int reverse_key;
         int dupsort;
         int create;
+        int integerkey;
+        int integerdup;
+        int dupfixed;
     } arg = {NULL, NULL, 0, 0, 1};
 
     static const struct argspec argspec[] = {
@@ -1551,6 +1569,9 @@ env_open_db(EnvObject *self, PyObject *args, PyObject *kwds)
         {"reverse_key", ARG_BOOL, OFFSET(env_open_db, reverse_key)},
         {"dupsort", ARG_BOOL, OFFSET(env_open_db, dupsort)},
         {"create", ARG_BOOL, OFFSET(env_open_db, create)},
+        {"integerkey", ARG_BOOL, OFFSET(env_open_db, integerkey)},
+        {"integerdup", ARG_BOOL, OFFSET(env_open_db, integerdup)},
+        {"dupfixed", ARG_BOOL, OFFSET(env_open_db, dupfixed)},
     };
     int flags;
 
@@ -1569,6 +1590,15 @@ env_open_db(EnvObject *self, PyObject *args, PyObject *kwds)
     if(arg.create) {
         flags |= MDB_CREATE;
     }
+    if(arg.integerkey) {
+        flags |= MDB_INTEGERKEY;
+    }
+    if(arg.integerdup) {
+        flags |= MDB_INTEGERDUP;
+    }
+    if(arg.dupfixed) {
+        flags |= MDB_DUPFIXED;
+    }
 
     if(arg.txn) {
         return (PyObject *) db_from_name(self, arg.txn->txn, arg.key, flags);
@@ -1688,6 +1718,34 @@ env_reader_check(EnvObject *self)
     return PyLong_FromLongLong(dead);
 }
 
+/**
+ * Environment.set_mapsize(size) -> None
+ */
+static PyObject *
+env_reader_set_mapsize(EnvObject *self, PyObject *args, PyObject *kwargs)
+{
+    struct env_set_mapsize {
+        size_t map_size;
+    } arg = {0};
+
+    static const struct argspec argspec[] = {
+        {"map_size", ARG_SIZE, OFFSET(env_set_mapsize, map_size)}
+    };
+    int rc;
+
+    static PyObject *cache = NULL;
+    if(parse_args(self->valid, SPECSIZE(), argspec, &cache,
+                  args, kwargs, &arg)) {
+        return NULL;
+    }
+
+    rc = mdb_env_set_mapsize(self->env, arg.map_size);
+    if(rc) {
+        return err_set("mdb_env_set_mapsize", rc);
+    }
+    Py_RETURN_NONE;
+}
+
 /**
  * Environment.sync()
  */
@@ -1738,8 +1796,8 @@ static struct PyMethodDef env_methods[] = {
     {"__exit__", (PyCFunction)env_exit, METH_VARARGS},
     {"begin", (PyCFunction)env_begin, METH_VARARGS|METH_KEYWORDS},
     {"close", (PyCFunction)env_close, METH_NOARGS},
-    {"copy", (PyCFunction)env_copy, METH_VARARGS},
-    {"copyfd", (PyCFunction)env_copyfd, METH_VARARGS},
+    {"copy", (PyCFunction)env_copy, METH_VARARGS|METH_KEYWORDS},
+    {"copyfd", (PyCFunction)env_copyfd, METH_VARARGS|METH_KEYWORDS},
     {"info", (PyCFunction)env_info, METH_NOARGS},
     {"flags", (PyCFunction)env_flags, METH_NOARGS},
     {"max_key_size", (PyCFunction)env_max_key_size, METH_NOARGS},
@@ -1749,6 +1807,8 @@ static struct PyMethodDef env_methods[] = {
     {"stat", (PyCFunction)env_stat, METH_NOARGS},
     {"readers", (PyCFunction)env_readers, METH_NOARGS},
     {"reader_check", (PyCFunction)env_reader_check, METH_NOARGS},
+    {"set_mapsize", (PyCFunction)env_reader_set_mapsize,
+     METH_VARARGS|METH_KEYWORDS},
     {"sync", (PyCFunction)env_sync, METH_VARARGS},
     {NULL, NULL}
 };
@@ -1883,7 +1943,11 @@ static int
 _cursor_get_c(CursorObject *self, enum MDB_cursor_op op)
 {
     int rc;
-    UNLOCKED(rc, mdb_cursor_get(self->curs, &self->key, &self->val, op));
+
+    Py_BEGIN_ALLOW_THREADS;
+    rc = mdb_cursor_get(self->curs, &self->key, &self->val, op);
+    Py_END_ALLOW_THREADS;
+
     self->positioned = rc == 0;
     self->last_mutation = self->trans->mutations;
     if(rc) {
@@ -2020,6 +2084,7 @@ cursor_item(CursorObject *self)
     PyObject *key;
     PyObject *val;
     PyObject *tup;
+    int rc = 0;
 
     if(! self->valid) {
         return err_invalid();
@@ -2032,6 +2097,7 @@ cursor_item(CursorObject *self)
 
     as_buffer = self->trans->flags & TRANS_BUFFERS;
     key = obj_from_val(&self->key, as_buffer);
+    PRELOAD_UNLOCKED(rc, self->val.mv_data, self->val.mv_size);
     val = obj_from_val(&self->val, as_buffer);
     tup = PyTuple_New(2);
     if(tup && key && val) {
@@ -2175,7 +2241,7 @@ cursor_put_multi(CursorObject *self, PyObject *args, PyObject *kwds)
         flags |= MDB_NOOVERWRITE;
     }
     if(arg.append) {
-        flags |= MDB_APPEND;
+        flags |= (self->trans->db->flags & MDB_DUPSORT) ? MDB_APPENDDUP : MDB_APPEND;
     }
 
     if(! ((iter = PyObject_GetIter(arg.items)))) {
@@ -2283,7 +2349,7 @@ cursor_put(CursorObject *self, PyObject *args, PyObject *kwds)
 static PyObject *
 do_cursor_replace(CursorObject *self, MDB_val *key, MDB_val *val)
 {
-    int rc;
+    int rc = 0;
     PyObject *old;
     MDB_val newval = *val;
 
@@ -2293,6 +2359,7 @@ do_cursor_replace(CursorObject *self, MDB_val *key, MDB_val *val)
             return NULL;
         }
         if(self->positioned) {
+            PRELOAD_UNLOCKED(rc, self->val.mv_data, self->val.mv_size);
             if(! ((old = obj_from_val(&self->val, 0)))) {
                 return NULL;
             }
@@ -2368,7 +2435,7 @@ cursor_pop(CursorObject *self, PyObject *args, PyObject *kwds)
         {"key", ARG_BUF, OFFSET(cursor_pop, key)},
     };
     PyObject *old;
-    int rc;
+    int rc = 0;
 
     static PyObject *cache = NULL;
     if(parse_args(self->valid, SPECSIZE(), argspec, &cache, args, kwds, &arg)) {
@@ -2382,6 +2449,7 @@ cursor_pop(CursorObject *self, PyObject *args, PyObject *kwds)
     if(! self->positioned) {
         Py_RETURN_NONE;
     }
+    PRELOAD_UNLOCKED(rc, self->val.mv_data, self->val.mv_size);
     if(! ((old = obj_from_val(&self->val, 0)))) {
         return NULL;
     }
@@ -2459,6 +2527,7 @@ cursor_set_range(CursorObject *self, PyObject *arg)
 static PyObject *
 cursor_set_range_dup(CursorObject *self, PyObject *args, PyObject *kwds)
 {
+    PyObject *ret;
     struct cursor_set_range_dup {
         MDB_val key;
         MDB_val value;
@@ -2473,9 +2542,17 @@ cursor_set_range_dup(CursorObject *self, PyObject *args, PyObject *kwds)
     if(parse_args(self->valid, SPECSIZE(), argspec, &cache, args, kwds, &arg)) {
         return NULL;
     }
+
     self->key = arg.key;
     self->val = arg.value;
-    return _cursor_get(self, MDB_GET_BOTH);
+    ret = _cursor_get(self, MDB_GET_BOTH_RANGE);
+
+    /* issue #126: MDB_GET_BOTH_RANGE does not satisfy its documentation, and
+     * fails to update `key` and `value` on success. Therefore explicitly call
+     * MDB_GET_CURRENT after MDB_GET_BOTH_RANGE. */
+    _cursor_get_c(self, MDB_GET_CURRENT);
+
+    return ret;
 }
 
 /**
@@ -2489,9 +2566,11 @@ cursor_value(CursorObject *self)
     }
     /* Must refresh `key` and `val` following mutation. */
     if(self->last_mutation != self->trans->mutations &&
-       _cursor_get_c(self, MDB_GET_CURRENT)) {
+        _cursor_get_c(self, MDB_GET_CURRENT)) {
         return NULL;
     }
+    PRELOAD_UNLOCKED(0, self->val.mv_data, self->val.mv_size);
+
     return obj_from_val(&self->val, self->trans->flags & TRANS_BUFFERS);
 }
 
@@ -2499,13 +2578,17 @@ static PyObject *
 new_iterator(CursorObject *cursor, IterValFunc val_func, MDB_cursor_op op)
 {
     IterObject *iter = PyObject_New(IterObject, &PyIterator_Type);
-    if(iter) {
-        iter->val_func = val_func;
-        iter->curs = cursor;
-        Py_INCREF(cursor);
-        iter->started = 0;
-        iter->op = op;
+    if (!iter) {
+        return NULL;
     }
+
+    iter->val_func = val_func;
+    iter->curs = cursor;
+    Py_INCREF(cursor);
+    iter->started = 0;
+    iter->op = op;
+
+    DEBUG("new_iterator: %p", (void *)iter)
     return (PyObject *) iter;
 }
 
@@ -2664,7 +2747,7 @@ static PyObject *cursor_enter(CursorObject *self)
 }
 
 /**
- * Transaction.__exit__()
+ * Cursor.__exit__()
  */
 static PyObject *cursor_exit(CursorObject *self, PyObject *args)
 {
@@ -2672,9 +2755,19 @@ static PyObject *cursor_exit(CursorObject *self, PyObject *args)
     Py_RETURN_NONE;
 }
 
+/**
+ * Cursor.close()
+ */
+static PyObject *cursor_close(CursorObject *self)
+{
+    cursor_clear(self);
+    Py_RETURN_NONE;
+}
+
 static struct PyMethodDef cursor_methods[] = {
     {"__enter__", (PyCFunction)cursor_enter, METH_NOARGS},
     {"__exit__", (PyCFunction)cursor_exit, METH_VARARGS},
+    {"close", (PyCFunction)cursor_close, METH_NOARGS},
     {"count", (PyCFunction)cursor_count, METH_NOARGS},
     {"delete", (PyCFunction)cursor_delete, METH_VARARGS|METH_KEYWORDS},
     {"first", (PyCFunction)cursor_first, METH_NOARGS},
@@ -2804,14 +2897,9 @@ iter_next(IterObject *self)
     return self->val_func(self->curs);
 }
 
-static struct PyMethodDef iter_methods[] = {
-    {"next", (PyCFunction)cursor_next, METH_NOARGS},
-    {NULL, NULL}
-};
-
 static PyTypeObject PyIterator_Type = {
     PyVarObject_HEAD_INIT(NULL, 0)
-    "Iterator",                 /*tp_name*/
+    "_Iterator",                 /*tp_name*/
     sizeof(IterObject),         /*tp_basicsize*/
     0,                          /*tp_itemsize*/
     (destructor) iter_dealloc,  /*tp_dealloc*/
@@ -2837,7 +2925,7 @@ static PyTypeObject PyIterator_Type = {
     0,                          /*tp_weaklistoffset*/
     (getiterfunc)iter_iter,     /*tp_iter*/
     (iternextfunc)iter_next,    /*tp_iternext*/
-    iter_methods,               /*tp_methods*/
+    0,                          /*tp_methods*/
     0,                          /*tp_members*/
     0,                          /*tp_getset*/
     0,                          /*tp_base*/
@@ -2888,6 +2976,11 @@ trans_clear(TransObject *self)
 static void
 trans_dealloc(TransObject *self)
 {
+    if(self->weaklist != NULL) {
+        MDEBUG("Clearing weaklist..")
+        PyObject_ClearWeakRefs((PyObject *) self);
+    }
+
     if(self->env && self->txn &&
        (self->env->max_spare_txns > 0) && (self->flags & TRANS_RDONLY)) {
         MDEBUG("caching trans")
@@ -2899,15 +2992,17 @@ trans_dealloc(TransObject *self)
         self->spare_next = self->env->spare_txns;
         self->env->spare_txns = self;
         self->env->max_spare_txns--;
+        Py_INCREF(self);
 
         Py_CLEAR(self->db);
         UNLINK_CHILD(self->env, self)
         Py_CLEAR(self->env);
-    } else {
-        MDEBUG("deleting trans")
-        trans_clear(self);
-        PyObject_Del(self);
+        return;
     }
+
+    MDEBUG("deleting trans")
+    trans_clear(self);
+    PyObject_Del(self);
 }
 
 /**
@@ -3131,7 +3226,11 @@ trans_get(TransObject *self, PyObject *args, PyObject *kwds)
         return type_error("key must be given.");
     }
 
-    UNLOCKED(rc, mdb_get(self->txn, arg.db->dbi, &arg.key, &val));
+    Py_BEGIN_ALLOW_THREADS
+    rc = mdb_get(self->txn, arg.db->dbi, &arg.key, &val);
+    preload(rc, val.mv_data, val.mv_size);
+    Py_END_ALLOW_THREADS
+
     if(rc) {
         if(rc == MDB_NOTFOUND) {
             Py_INCREF(arg.default_);
@@ -3267,7 +3366,7 @@ trans_pop(TransObject *self, PyObject *args, PyObject *kwds)
     };
     CursorObject *cursor;
     PyObject *old;
-    int rc;
+    int rc = 0;
 
     static PyObject *cache = NULL;
     if(parse_args(self->valid, SPECSIZE(), argspec, &cache, args, kwds, &arg)) {
@@ -3290,6 +3389,8 @@ trans_pop(TransObject *self, PyObject *args, PyObject *kwds)
         Py_DECREF((PyObject *)cursor);
         Py_RETURN_NONE;
     }
+
+    PRELOAD_UNLOCKED(rc, cursor->val.mv_data, cursor->val.mv_size);
     if(! ((old = obj_from_val(&cursor->val, 0)))) {
         Py_DECREF((PyObject *)cursor);
         return NULL;
@@ -3332,6 +3433,21 @@ static PyObject *trans_exit(TransObject *self, PyObject *args)
     }
 }
 
+/**
+ * Transaction.id() -> int
+ */
+static PyObject *trans_id(TransObject *self)
+{
+    size_t id;
+
+    if(! self->valid) {
+        return err_invalid();
+    }
+
+    id = mdb_txn_id(self->txn);
+    return PyLong_FromUnsignedLong(id);
+}
+
 /**
  * Transaction.stat() -> dict
  */
@@ -3375,6 +3491,7 @@ static struct PyMethodDef trans_methods[] = {
     {"put", (PyCFunction)trans_put, METH_VARARGS|METH_KEYWORDS},
     {"replace", (PyCFunction)trans_replace, METH_VARARGS|METH_KEYWORDS},
     {"pop", (PyCFunction)trans_pop, METH_VARARGS|METH_KEYWORDS},
+    {"id", (PyCFunction)trans_id, METH_NOARGS},
     {"stat", (PyCFunction)trans_stat, METH_VARARGS|METH_KEYWORDS},
     {NULL, NULL}
 };
@@ -3404,7 +3521,7 @@ static PyTypeObject PyTransaction_Type = {
     0,                          /*tp_traverse*/
     (inquiry) trans_clear,      /*tp_clear*/
     0,                          /*tp_richcompare*/
-    0,                          /*tp_weaklistoffset*/
+    offsetof(TransObject, weaklist), /*tp_weaklistoffset*/
     0,                          /*tp_iter*/
     0,                          /*tp_iternext*/
     trans_methods,              /*tp_methods*/
@@ -3422,6 +3539,31 @@ static PyTypeObject PyTransaction_Type = {
 };
 
 
+/**
+ * Construct a PyString and append it to a list, returning 0 on success or -1
+ * on error.
+ */
+static int
+append_string(PyObject *list, const char *s)
+{
+#if PY_MAJOR_VERSION >= 3
+    PyObject *o = PyUnicode_FromString(s);
+#else
+    PyObject *o = PyBytes_FromStringAndSize(s, strlen(s));
+#endif
+
+    if(! o) {
+        return -1;
+    }
+    if(PyList_Append(list, o)) {
+        Py_DECREF(o);
+        return -1;
+    }
+    Py_DECREF(o);
+    return 0;
+}
+
+
 /**
  * lmdb.enable_drop_gil()
  */
@@ -3464,7 +3606,7 @@ static struct PyModuleDef moduledef = {
 /**
  * Initialize and publish the LMDB built-in types.
  */
-static int init_types(PyObject *mod)
+static int init_types(PyObject *mod, PyObject *__all__)
 {
     static PyTypeObject *types[] = {
         &PyEnvironment_Type,
@@ -3478,13 +3620,22 @@ static int init_types(PyObject *mod)
     int i;
     for(i = 0; types[i]; i++) {
         PyTypeObject *type = types[i];
+        char const * name = type->tp_name;
+
         if(PyType_Ready(type)) {
             return -1;
         }
-        if(PyObject_SetAttrString(mod, type->tp_name, (PyObject *)type)) {
+        if(PyObject_SetAttrString(mod, name, (PyObject *)type)) {
+            return -1;
+        }
+
+        /* As a special case, export _Database */
+        if((name[0] != '_' || 0 == strcmp(name, "_Database")) &&
+           append_string(__all__, name)) {
             return -1;
         }
     }
+
     return 0;
 }
 
@@ -3509,11 +3660,11 @@ static int init_constants(PyObject *mod)
  * Create lmdb.Error exception class, and one subclass for each entry in
  * `error_map`.
  */
-static int init_errors(PyObject *mod)
+static int init_errors(PyObject *mod, PyObject *__all__)
 {
     size_t count;
     char qualname[64];
-    int i;
+    size_t i;
 
     Error = PyErr_NewException("lmdb.Error", NULL, NULL);
     if(! Error) {
@@ -3522,6 +3673,9 @@ static int init_errors(PyObject *mod)
     if(PyObject_SetAttrString(mod, "Error", Error)) {
         return -1;
     }
+    if(append_string(__all__, "Error")) {
+        return -1;
+    }
 
     count = (sizeof error_map / sizeof error_map[0]);
     error_tbl = malloc(sizeof(PyObject *) * count);
@@ -3544,6 +3698,9 @@ static int init_errors(PyObject *mod)
         if(PyObject_SetAttrString(mod, error->name, klass)) {
             return -1;
         }
+        if(append_string(__all__, error->name)) {
+            return -1;
+        }
     }
     return 0;
 }
@@ -3554,6 +3711,7 @@ static int init_errors(PyObject *mod)
 PyMODINIT_FUNC
 MODINIT_NAME(void)
 {
+    PyObject *__all__;
 #if PY_MAJOR_VERSION >= 3
     PyObject *mod = PyModule_Create(&moduledef);
 #else
@@ -3563,7 +3721,17 @@ MODINIT_NAME(void)
         MOD_RETURN(NULL);
     }
 
-    if(init_types(mod)) {
+    if(! ((__all__ = PyList_New(0)))) {
+        MOD_RETURN(NULL);
+    }
+
+    if(init_types(mod, __all__)) {
+        MOD_RETURN(NULL);
+    }
+    if(append_string(__all__, "enable_drop_gil")) {
+        MOD_RETURN(NULL);
+    }
+    if(append_string(__all__, "version")) {
         MOD_RETURN(NULL);
     }
 
@@ -3583,11 +3751,17 @@ MODINIT_NAME(void)
     if(init_constants(mod)) {
         MOD_RETURN(NULL);
     }
-    if(init_errors(mod)) {
+    if(init_errors(mod, __all__)) {
         MOD_RETURN(NULL);
     }
     if(PyObject_SetAttrString(mod, "open", (PyObject *)&PyEnvironment_Type)) {
         MOD_RETURN(NULL);
     }
+
+    if(PyObject_SetAttrString(mod, "__all__", __all__)) {
+        MOD_RETURN(NULL);
+    }
+    Py_DECREF(__all__);
+
     MOD_RETURN(mod);
 }
